Ad

Struct Ad 

pub struct Ad<const N: usize> { /* private fields */ }
Automatic differentiation value tracking first and second derivatives

Value getters:

  • value() -> f64: Returns the current numerical value
  • grad() -> SVector<f64, N>: Returns the gradient vector
  • hess() -> SMatrix<f64, N, N>: Returns the Hessian matrix

Type Parameters

  • N - The dimension of the input space (number of variables)

Fields (private)

  • value - The current value of the function
  • grad - The gradient (first derivatives) as a vector
  • hess - The Hessian matrix (second derivatives)

Implementations

Source

impl<const N: usize> Ad<N>

pub fn neg(&self) -> Self

pub fn sqrt(&self) -> Self

pub fn square(&self) -> Self

pub fn powi(&self, exponent: i32) -> Self

Examples found in repository?
src/examples/mass_spring.rs (line 20)
14    fn eval(&self, variables: &raddy::types::advec<4, 4>, _: &()) -> raddy::Ad<4> {
15        let p1 = advec::<4, 2>::new(variables[0].clone(), variables[1].clone());
16        let p2 = advec::<4, 2>::new(variables[2].clone(), variables[3].clone());
17
18        let len = (p2 - p1).norm();
19        // Hooke's law
20        let potential = val::scalar(0.5 * self.k) * (len - val::scalar(self.restlen)).powi(2);
21
22        potential
23    }
pub fn powf(&self, exponent: f64) -> Self

pub fn abs(&self) -> Self

pub fn exp(&self) -> Self

pub fn ln(&self) -> Self

Examples found in repository?
src/examples/basic.rs (line 17)
9fn main() {
10    // 1.
11    // ################ scalar ################
12    let mut rng = thread_rng();
13    let val = rng.gen_range(0.0..10.0);
14
15    let var = var::scalar(val);
16    let var = &var;
17    let y = var.sin() * var + var.ln();
18    let g = val * val.cos() + val.sin() + val.recip();
19    let h = -val * val.sin() + 2.0 * val.cos() - val.powi(-2);
20
21    assert_abs_diff_eq!(y.grad()[(0, 0)], g, epsilon = EPS);
22    assert_abs_diff_eq!(y.hess()[(0, 0)], h, epsilon = EPS);
23
24    // 2.
25    // ############################# Matrix #############################
26
27    const N_TEST_MAT_4: usize = 4;
28    type NaConst = Const<N_TEST_MAT_4>;
29    const N_VEC_4: usize = N_TEST_MAT_4 * N_TEST_MAT_4;
30
31    let vals: &[f64] = &(0..N_VEC_4)
32        .map(|_| rng.gen_range(-4.0..4.0))
33        .collect::<Vec<_>>();
34
35    let s: SVector<Ad<N_VEC_4>, N_VEC_4> = var::vector_from_slice(vals);
36    let z = s
37        .clone()
38        // This reshape is COL MAJOR!!!!!!!!!!!!!
39        .reshape_generic(NaConst {}, NaConst {})
40        .transpose();
41
42    let det = z.determinant();
43    let _grad = det.grad();
44    let _hess = det.hess();
45    // core logic ends ####################################################
46
47    // correctness
48    // let expected_grad = grad_det4(
49    //     vals[0], vals[1], vals[2], vals[3], vals[4], vals[5], vals[6], vals[7], vals[8], vals[9],
50    //     vals[10], vals[11], vals[12], vals[13], vals[14], vals[15],
51    // );
52    // let g_diff = (expected_grad - det.grad()).norm_squared();
53    // assert_abs_diff_eq!(g_diff, 0.0, epsilon = EPS);
54
55    // let expected_hess = hess_det4(
56    //     vals[0], vals[1], vals[2], vals[3], vals[4], vals[5], vals[6], vals[7], vals[8], vals[9],
57    //     vals[10], vals[11], vals[12], vals[13], vals[14], vals[15],
58    // );
59    // let h_diff = (det.hess() - expected_hess).norm_squared();
60    // assert_abs_diff_eq!(h_diff, 0.0, epsilon = EPS);
61}
pub fn log(&self, base: f64) -> Self

pub fn log2(&self) -> Self

pub fn log10(&self) -> Self

pub fn sin(&self) -> Self

Examples found in repository?
src/examples/basic.rs (line 17)
9fn main() {
10    // 1.
11    // ################ scalar ################
12    let mut rng = thread_rng();
13    let val = rng.gen_range(0.0..10.0);
14
15    let var = var::scalar(val);
16    let var = &var;
17    let y = var.sin() * var + var.ln();
18    let g = val * val.cos() + val.sin() + val.recip();
19    let h = -val * val.sin() + 2.0 * val.cos() - val.powi(-2);
20
21    assert_abs_diff_eq!(y.grad()[(0, 0)], g, epsilon = EPS);
22    assert_abs_diff_eq!(y.hess()[(0, 0)], h, epsilon = EPS);
23
24    // 2.
25    // ############################# Matrix #############################
26
27    const N_TEST_MAT_4: usize = 4;
28    type NaConst = Const<N_TEST_MAT_4>;
29    const N_VEC_4: usize = N_TEST_MAT_4 * N_TEST_MAT_4;
30
31    let vals: &[f64] = &(0..N_VEC_4)
32        .map(|_| rng.gen_range(-4.0..4.0))
33        .collect::<Vec<_>>();
34
35    let s: SVector<Ad<N_VEC_4>, N_VEC_4> = var::vector_from_slice(vals);
36    let z = s
37        .clone()
38        // This reshape is COL MAJOR!!!!!!!!!!!!!
39        .reshape_generic(NaConst {}, NaConst {})
40        .transpose();
41
42    let det = z.determinant();
43    let _grad = det.grad();
44    let _hess = det.hess();
45    // core logic ends ####################################################
46
47    // correctness
48    // let expected_grad = grad_det4(
49    //     vals[0], vals[1], vals[2], vals[3], vals[4], vals[5], vals[6], vals[7], vals[8], vals[9],
50    //     vals[10], vals[11], vals[12], vals[13], vals[14], vals[15],
51    // );
52    // let g_diff = (expected_grad - det.grad()).norm_squared();
53    // assert_abs_diff_eq!(g_diff, 0.0, epsilon = EPS);
54
55    // let expected_hess = hess_det4(
56    //     vals[0], vals[1], vals[2], vals[3], vals[4], vals[5], vals[6], vals[7], vals[8], vals[9],
57    //     vals[10], vals[11], vals[12], vals[13], vals[14], vals[15],
58    // );
59    // let h_diff = (det.hess() - expected_hess).norm_squared();
60    // assert_abs_diff_eq!(h_diff, 0.0, epsilon = EPS);
61}
pub fn cos(&self) -> Self

pub fn tan(&self) -> Self

pub fn asin(&self) -> Self

pub fn acos(&self) -> Self

pub fn atan(&self) -> Self

👎Deprecated: Please use atan2 instead.
pub fn sinh(&self) -> Self

pub fn cosh(&self) -> Self

pub fn tanh(&self) -> Self

pub fn asinh(&self) -> Self

pub fn acosh(&self) -> Self

pub fn atanh(&self) -> Self

Source

impl<const N: usize> Ad<N>

pub fn add_value(&self, other: f64) -> Self

pub fn sub_value(&self, other: f64) -> Self

pub fn mul_value(&self, other: f64) -> Self

pub fn recip(&self) -> Self

pub fn div_value(&self, other: f64) -> Self

pub fn atan2(&self, x: &Self) -> Self

self is y
pub fn min(&self, other: &Self) -> Self

pub fn max(&self, other: &Self) -> Self

pub fn clamp(&self, low: &Self, high: &Self) -> Self

pub fn hypot(&self, other: &Self) -> Self

Source

impl<const N: usize> Ad<N>

pub fn value(&self) -> f64

Returns the current value of the AD variable

pub fn grad(&self) -> SVector<f64, L>

Returns the gradient (first derivatives) of the AD variable

Returns

The gradient (A vector containing the partial derivatives with respect to each input variable).

Examples found in repository?
src/examples/basic.rs (line 21)
9fn main() {
10    // 1.
11    // ################ scalar ################
12    let mut rng = thread_rng();
13    let val = rng.gen_range(0.0..10.0);
14
15    let var = var::scalar(val);
16    let var = &var;
17    let y = var.sin() * var + var.ln();
18    let g = val * val.cos() + val.sin() + val.recip();
19    let h = -val * val.sin() + 2.0 * val.cos() - val.powi(-2);
20
21    assert_abs_diff_eq!(y.grad()[(0, 0)], g, epsilon = EPS);
22    assert_abs_diff_eq!(y.hess()[(0, 0)], h, epsilon = EPS);
23
24    // 2.
25    // ############################# Matrix #############################
26
27    const N_TEST_MAT_4: usize = 4;
28    type NaConst = Const<N_TEST_MAT_4>;
29    const N_VEC_4: usize = N_TEST_MAT_4 * N_TEST_MAT_4;
30
31    let vals: &[f64] = &(0..N_VEC_4)
32        .map(|_| rng.gen_range(-4.0..4.0))
33        .collect::<Vec<_>>();
34
35    let s: SVector<Ad<N_VEC_4>, N_VEC_4> = var::vector_from_slice(vals);
36    let z = s
37        .clone()
38        // This reshape is COL MAJOR!!!!!!!!!!!!!
39        .reshape_generic(NaConst {}, NaConst {})
40        .transpose();
41
42    let det = z.determinant();
43    let _grad = det.grad();
44    let _hess = det.hess();
45    // core logic ends ####################################################
46
47    // correctness
48    // let expected_grad = grad_det4(
49    //     vals[0], vals[1], vals[2], vals[3], vals[4], vals[5], vals[6], vals[7], vals[8], vals[9],
50    //     vals[10], vals[11], vals[12], vals[13], vals[14], vals[15],
51    // );
52    // let g_diff = (expected_grad - det.grad()).norm_squared();
53    // assert_abs_diff_eq!(g_diff, 0.0, epsilon = EPS);
54
55    // let expected_hess = hess_det4(
56    //     vals[0], vals[1], vals[2], vals[3], vals[4], vals[5], vals[6], vals[7], vals[8], vals[9],
57    //     vals[10], vals[11], vals[12], vals[13], vals[14], vals[15],
58    // );
59    // let h_diff = (det.hess() - expected_hess).norm_squared();
60    // assert_abs_diff_eq!(h_diff, 0.0, epsilon = EPS);
61}
pub fn hess(&self) -> SMatrix<f64, RC, RC>

Returns the Hessian matrix (second derivatives) of the AD variable

Returns

The hessian.

Examples found in repository?
src/examples/basic.rs (line 22)
9fn main() {
10    // 1.
11    // ################ scalar ################
12    let mut rng = thread_rng();
13    let val = rng.gen_range(0.0..10.0);
14
15    let var = var::scalar(val);
16    let var = &var;
17    let y = var.sin() * var + var.ln();
18    let g = val * val.cos() + val.sin() + val.recip();
19    let h = -val * val.sin() + 2.0 * val.cos() - val.powi(-2);
20
21    assert_abs_diff_eq!(y.grad()[(0, 0)], g, epsilon = EPS);
22    assert_abs_diff_eq!(y.hess()[(0, 0)], h, epsilon = EPS);
23
24    // 2.
25    // ############################# Matrix #############################
26
27    const N_TEST_MAT_4: usize = 4;
28    type NaConst = Const<N_TEST_MAT_4>;
29    const N_VEC_4: usize = N_TEST_MAT_4 * N_TEST_MAT_4;
30
31    let vals: &[f64] = &(0..N_VEC_4)
32        .map(|_| rng.gen_range(-4.0..4.0))
33        .collect::<Vec<_>>();
34
35    let s: SVector<Ad<N_VEC_4>, N_VEC_4> = var::vector_from_slice(vals);
36    let z = s
37        .clone()
38        // This reshape is COL MAJOR!!!!!!!!!!!!!
39        .reshape_generic(NaConst {}, NaConst {})
40        .transpose();
41
42    let det = z.determinant();
43    let _grad = det.grad();
44    let _hess = det.hess();
45    // core logic ends ####################################################
46
47    // correctness
48    // let expected_grad = grad_det4(
49    //     vals[0], vals[1], vals[2], vals[3], vals[4], vals[5], vals[6], vals[7], vals[8], vals[9],
50    //     vals[10], vals[11], vals[12], vals[13], vals[14], vals[15],
51    // );
52    // let g_diff = (expected_grad - det.grad()).norm_squared();
53    // assert_abs_diff_eq!(g_diff, 0.0, epsilon = EPS);
54
55    // let expected_hess = hess_det4(
56    //     vals[0], vals[1], vals[2], vals[3], vals[4], vals[5], vals[6], vals[7], vals[8], vals[9],
57    //     vals[10], vals[11], vals[12], vals[13], vals[14], vals[15],
58    // );
59    // let h_diff = (det.hess() - expected_hess).norm_squared();
60    // assert_abs_diff_eq!(h_diff, 0.0, epsilon = EPS);
61}
Source

impl Ad<1>

pub fn given_scalar(value: f64, grad: f64, hess: f64) -> Self

Creates an AD scalar with explicitly specified value, gradient and Hessian

Arguments
  • value - The scalar value
  • grad - The gradient (first derivative)
  • hess - The Hessian (second derivative)
Returns

A new Ad<1> instance with the specified properties

pub fn active_scalar(value: f64) -> Self

Creates an active scalar AD value with unit gradient

Arguments
  • value - The scalar value
Returns

A new Ad<1> instance that is active (gradient = 1.0)

Source

impl<const N: usize> Ad<N>

pub fn inactive_scalar(value: f64) -> Self

Creates an inactive scalar AD value with zero gradient and Hessian

Arguments
  • value - The scalar value
Returns

A new Ad<N> instance that is inactive (gradient = 0)

pub fn inactive_vector<const L: usize>( values: &SVector<f64, N>, ) -> SVector<Self, L>

Creates a vector of inactive AD values from a vector of f64 values

Arguments
  • values - Input vector of numerical values
Type Parameters
  • L - Length of the output vector
Returns

A vector of inactive AD values

pub fn inactive_from_slice<const L: usize>(values: &[f64]) -> SVector<Self, L>

Creates a vector of inactive AD values from a slice of f64 values

Arguments
  • values - Slice of numerical values
Type Parameters
  • L - Length of the output vector
Returns

A vector of inactive AD values

Panics

If the slice length doesn’t match the input dimension N

pub fn given_vector( value: f64, grad: &SVector<f64, L>, hess: &SMatrix<f64, RC, RC>, ) -> Self

Creates an AD value with explicitly specified value, gradient and Hessian

Arguments
  • value - The scalar value
  • grad - The gradient vector
  • hess - The Hessian matrix
Returns

A new Ad<N> instance with the specified properties

pub fn active_vector(vector: &SVector<f64, N>) -> SVector<Self, N>

Creates a vector of active AD values from a vector of f64 values

Arguments
  • values - Input vector of numerical values
Returns

A vector of active AD values where each element has unit gradient in its corresponding dimension

pub fn active_from_slice(values: &[f64]) -> SVector<Self, N>

Creates a vector of active AD values from a slice of f64 values

Arguments
  • values - Slice of numerical values
Returns

A vector of active AD values

Panics

If the slice length doesn’t match the input dimension N